@@ -20,6 +20,7 @@ from typing import Dict, List, Tuple, Union
 from multiprocessing import cpu_count
 from sklearn import random_projection
 from sklearn.decomposition import KernelPCA
+from sklearn.impute import SimpleImputer
 from sklearn.manifold import TSNE
 from sklearn.preprocessing import MinMaxScaler, StandardScaler, LabelEncoder
 from tqdm import tqdm
@@ -55,6 +56,10 @@ class project:
         arena_dims: tuple = (1,),
         exclude_bodyparts: List = tuple([""]),
         exp_conditions: dict = None,
+        interpolate_outliers: bool = True,
+        interpolation_limit: int = 5,
+        interpolation_std: int = 5,
+        likelihood_tol: float = 0.75,
         model: str = "mouse_topview",
         path: str = deepof.utils.os.path.join("."),
         smooth_alpha: float = 0.99,
@@ -68,7 +73,7 @@ class project:
 
         self.table_format = table_format
         if self.table_format == "autodetect":
-            ex = os.listdir(self.table_path)[0]
+            ex = [i for i in os.listdir(self.table_path) if not i.startswith(".")][0]
             if ".h5" in ex:
                 self.table_format = ".h5"
             elif ".csv" in ex:
@@ -88,17 +93,21 @@ class project:
                 if tab.endswith(self.table_format) and not tab.startswith(".")
             ]
         )
+        self.angles = True
         self.animal_ids = animal_ids
         self.arena = arena
         self.arena_dims = arena_dims
-        self.exp_conditions = exp_conditions
-        self.scales = self.get_scale
-        self.smooth_alpha = smooth_alpha
-        self.video_format = video_format
-        self.angles = True
         self.distances = "all"
         self.ego = False
+        self.exp_conditions = exp_conditions
+        self.interpolate_outliers = interpolate_outliers
+        self.interpolation_limit = interpolation_limit
+        self.interpolation_std = interpolation_std
+        self.likelihood_tolerance = likelihood_tol
+        self.scales = self.get_scale
+        self.smooth_alpha = smooth_alpha
         self.subset_condition = None
+        self.video_format = video_format
 
         model_dict = {
             "mouse_topview": deepof.utils.connect_mouse_topview(animal_ids[0])
@@ -149,16 +158,16 @@ class project:
 
             scales = []
             for vid_index, _ in enumerate(self.videos):
+
+                ellipse = deepof.utils.recognize_arena(
+                    self.videos,
+                    vid_index,
+                    path=self.video_path,
+                    arena_type=self.arena,
+                )[0]
+
                 scales.append(
-                    list(
-                        deepof.utils.recognize_arena(
-                            self.videos,
-                            vid_index,
-                            path=self.video_path,
-                            arena_type=self.arena,
-                        )[0]
-                        * 2
-                    )
+                    list(np.array([ellipse[0][0], ellipse[0][1], ellipse[1][1]]) * 2)
                     + list(self.arena_dims)
                 )
 
@@ -190,37 +199,25 @@ class project:
             }
         elif self.table_format == ".csv":
 
-            for tab in self.tables:
-                head = pd.read_csv(
-                    deepof.utils.os.path.join(self.table_path, tab), nrows=2
-                )
-                data = pd.read_csv(
+            tab_dict = {
+                deepof.utils.re.findall("(.*)DLC", tab)[0]: pd.read_csv(
                     deepof.utils.os.path.join(self.table_path, tab),
-                    skiprows=2,
-                    index_col="coords",
-                    dtype={"coords": int},
+                    header=[0, 1, 2],
+                    index_col=0,
+                    dtype=float,
                 )
-                data.columns = pd.MultiIndex.from_product(
-                    [
-                        [head.columns[2]],
-                        set(list(head.iloc[0])[2:]),
-                        ["x", "y", "likelihood"],
-                    ],
-                    names=["scorer", "bodyparts", "coords"],
-                )
-                tab_dict[deepof.utils.re.findall("(.*)DLC", tab)[0]] = data
+                for tab in self.tables
+            }
 
         lik_dict = defaultdict()
 
         for key, value in tab_dict.items():
             x = value.xs("x", level="coords", axis=1, drop_level=False)
             y = value.xs("y", level="coords", axis=1, drop_level=False)
-            lik: pd.DataFrame = value.xs(
-                "likelihood", level="coords", axis=1, drop_level=True
-            )
+            lik = value.xs("likelihood", level="coords", axis=1, drop_level=True)
 
             tab_dict[key] = pd.concat([x, y], axis=1).sort_index(axis=1)
-            lik_dict[key] = lik
+            lik_dict[key] = lik.droplevel("scorer", axis=1)
 
         if self.smooth_alpha:
 
@@ -235,7 +232,7 @@ class project:
                     )
                 )
                 smooth.columns = cols
-                tab_dict[key] = smooth
+                tab_dict[key] = smooth.iloc[1:, :].reset_index(drop=True)
 
         for key, tab in tab_dict.items():
             tab_dict[key] = tab[tab.columns.levels[0][0]]
@@ -270,6 +267,21 @@ class project:
                 )
                 tab_dict[k] = temp.sort_index(axis=1)
 
+        if self.interpolate_outliers:
+
+            if verbose:
+                print("Interpolating outliers...")
+
+            for k, value in tab_dict.items():
+                tab_dict[k] = deepof.utils.interpolate_outliers(
+                    value,
+                    lik_dict[k],
+                    likelihood_tolerance=self.likelihood_tolerance,
+                    mode="or",
+                    limit=self.interpolation_limit,
+                    n_std=self.interpolation_std,
+                )
+
         return tab_dict, lik_dict
 
     def get_distances(self, tab_dict: dict, verbose: bool = False) -> dict:
@@ -532,7 +544,7 @@ class coordinates:
 
         if speed:
             for key, tab in tabs.items():
-                vel = deepof.utils.rolling_speed(tab, deriv=speed + 1, center=center)
+                vel = deepof.utils.rolling_speed(tab, deriv=speed, center=center)
                 tabs[key] = vel
 
         if length:
@@ -696,7 +708,11 @@ class coordinates:
 
     # noinspection PyDefaultArgument
     def rule_based_annotation(
-        self, hparams: Dict = {}, video_output: bool = False, frame_limit: int = np.inf
+        self,
+        params: Dict = {},
+        video_output: bool = False,
+        frame_limit: int = np.inf,
+        debug: bool = False,
     ) -> Table_dict:
         """Annotates coordinates using a simple rule-based pipeline"""
 
@@ -720,7 +736,7 @@ class coordinates:
                 arena_type=self._arena,
                 recog_limit=1,
                 path=os.path.join(self._path, "Videos"),
-                hparams=hparams,
+                params=params,
             )
 
         if video_output:  # pragma: no cover
@@ -734,10 +750,11 @@ class coordinates:
                     self._videos,
                     list(self._tables.keys()).index(idx),
                     tag_dict[idx],
+                    debug=debug,
                     frame_limit=frame_limit,
                     recog_limit=1,
                     path=os.path.join(self._path, "Videos"),
-                    hparams=hparams,
+                    params=params,
                 )
                 pbar.update(1)
 
@@ -760,6 +777,9 @@ class coordinates:
             tag_dict, typ="rule-based", arena=self._arena, arena_dims=self._arena_dims
         )
 
+    def gmvae_embedding(self):
+        pass
+
 
 class table_dict(dict):
     """
@@ -945,8 +965,8 @@ class table_dict(dict):
             ).reshape(X_train.shape)
 
             if scale == "standard":
-                assert np.allclose(np.mean(X_train), 0)
-                assert np.allclose(np.std(X_train), 1)
+                assert np.allclose(np.nan_to_num(np.mean(X_train), nan=0), 0)
+                assert np.allclose(np.nan_to_num(np.std(X_train), nan=1), 1)
 
             if test_videos:
                 X_test = scaler.transform(X_test.reshape(-1, X_test.shape[-1])).reshape(
@@ -1030,6 +1050,7 @@ class table_dict(dict):
 
         # noinspection PyUnresolvedReferences
         X = X[np.random.choice(X.shape[0], sample, replace=False), :]
+        X = SimpleImputer(strategy="median").fit_transform(X)
 
         rproj = random_projection.GaussianRandomProjection(n_components=n_components)
         X = rproj.fit_transform(X)
@@ -1044,6 +1065,7 @@ class table_dict(dict):
         performance or visualization reasons"""
 
         X = self.get_training_set()[0]
+        X = SimpleImputer(strategy="median").fit_transform(X)
 
         # Takes care of propagated labels if present
         if self._propagate_labels:
@@ -1065,6 +1087,7 @@ class table_dict(dict):
         performance or visualization reasons"""
 
         X = self.get_training_set()[0]
+        X = SimpleImputer(strategy="median").fit_transform(X)
 
         # Takes care of propagated labels if present
         if self._propagate_labels:
@@ -1108,3 +1131,5 @@ def merge_tables(*args):
 #   - with the current implementation, preprocess can't fully work on merged table_dict instances.
 #   While some operations (mainly alignment) should be carried out before merging, others require
 #   the whole dataset to function properly.
+#   - Understand how keras handles NA values. Decide whether to do nothing, to mask them or to impute them using
+#   a clear outlier (e.g. -9999)
