@@ -16,10 +16,15 @@ import os
 import pandas as pd
 import regex as re
 import seaborn as sns
+import warnings
 from itertools import combinations
 from scipy import stats
 from typing import Any, List, NewType
 
+# Ignore warning with no downstream effect
+warnings.filterwarnings("ignore", message="All-NaN slice encountered")
+
+# Create custom string type
 Coordinates = NewType("Coordinates", Any)
 
 
@@ -133,10 +138,34 @@ def climb_wall(
 
     nose = pos_dict[nose]
 
+    def rotate(origin, point, angle):
+        ox, oy = origin
+        px, py = point
+
+        qx = ox + np.cos(angle) * (px - ox) - np.sin(angle) * (py - oy)
+        qy = oy + np.sin(angle) * (px - ox) + np.cos(angle) * (py - oy)
+        return qx, qy
+
+    def outside_ellipse(x, y, e_center, e_axes, e_angle, threshold=0.0):
+
+        x, y = rotate(e_center, (x, y), np.radians(e_angle))
+
+        term_x = (x - e_center[0]) ** 2 / (e_axes[0] + threshold) ** 2
+        term_y = (y - e_center[1]) ** 2 / (e_axes[1] + threshold) ** 2
+        return term_x + term_y > 1
+
     if arena_type == "circular":
-        center = np.zeros(2) if centered_data else np.array(arena[:2])
-        radius = arena[2]
-        climbing = np.linalg.norm(nose - center, axis=1) > (radius + tol)
+        center = np.zeros(2) if centered_data else np.array(arena[0])
+        axes = arena[1]
+        angle = arena[2]
+        climbing = outside_ellipse(
+            x=nose["x"],
+            y=nose["y"],
+            e_center=center,
+            e_axes=axes,
+            e_angle=-angle,
+            threshold=tol,
+        )
 
     else:
         raise NotImplementedError("Supported values for arena_type are ['circular']")
@@ -148,7 +177,6 @@ def huddle(
     pos_dframe: pd.DataFrame,
     speed_dframe: pd.DataFrame,
     tol_forward: float,
-    tol_spine: float,
     tol_speed: float,
     animal_id: str = "",
 ) -> np.array:
@@ -173,34 +201,20 @@ def huddle(
 
     forward = (
         np.linalg.norm(
-            pos_dframe[animal_id + "Left_ear"] - pos_dframe[animal_id + "Left_fhip"],
+            pos_dframe[animal_id + "Left_bhip"] - pos_dframe[animal_id + "Left_fhip"],
             axis=1,
         )
         < tol_forward
-    ) & (
+    ) | (
         np.linalg.norm(
-            pos_dframe[animal_id + "Right_ear"] - pos_dframe[animal_id + "Right_fhip"],
+            pos_dframe[animal_id + "Right_bhip"] - pos_dframe[animal_id + "Right_fhip"],
             axis=1,
         )
         < tol_forward
     )
 
-    spine = [
-        animal_id + "Spine_1",
-        animal_id + "Center",
-        animal_id + "Spine_2",
-        animal_id + "Tail_base",
-    ]
-    spine_dists = []
-    for comb in range(2):
-        spine_dists.append(
-            np.linalg.norm(
-                pos_dframe[spine[comb]] - pos_dframe[spine[comb + 1]], axis=1
-            )
-        )
-    spine = np.mean(spine_dists) < tol_spine
     speed = speed_dframe[animal_id + "Center"] < tol_speed
-    hudd = forward & spine & speed
+    hudd = forward & speed
 
     return hudd
 
@@ -376,18 +390,18 @@ def get_hparameters(hparams: dict = {}) -> dict:
         - hparams (dict): dictionary containing hyperparameters to overwrite
 
     Returns:
-        - defaults (dict): dictionary with overwriten parameters. Those not
+        - defaults (dict): dictionary with overwritten parameters. Those not
         specified in the input retain their default values"""
 
     defaults = {
-        "speed_pause": 3,
+        "speed_pause": 5,
+        "climb_tol": 10,
         "close_contact_tol": 35,
         "side_contact_tol": 80,
         "follow_frames": 10,
         "follow_tol": 5,
         "huddle_forward": 15,
-        "huddle_spine": 10,
-        "huddle_speed": 0.1,
+        "huddle_speed": 1,
         "fps": 24,
     }
 
@@ -435,7 +449,7 @@ def rule_based_tagging(
     arena_type: str,
     recog_limit: int = 1,
     path: str = os.path.join("."),
-    hparams: dict = {},
+    params: dict = {},
 ) -> pd.DataFrame:
     """Outputs a dataframe with the registered motives per frame. If specified, produces a labeled
     video displaying the information in real time
@@ -450,7 +464,7 @@ def rule_based_tagging(
         - vid_index (int): index in videos of the experiment to annotate
         - path (str): directory in which the experimental data is stored
         - recog_limit (int): number of frames to use for arena recognition (1 by default)
-        - hparams (dict): dictionary to overwrite the default values of the hyperparameters of the functions
+        - params (dict): dictionary to overwrite the default values of the hyperparameters of the functions
         that the rule-based pose estimation utilizes. Values can be:
             - speed_pause (int): size of the rolling window to use when computing speeds
             - close_contact_tol (int): maximum distance between single bodyparts that can be used to report the trait
@@ -459,14 +473,13 @@ def rule_based_tagging(
             - follow_tol (int): maximum distance between follower and followed's path during the last follow_frames,
             in order to report a detection
             - huddle_forward (int): maximum distance between ears and forward limbs to report a huddle detection
-            - huddle_spine (int): maximum average distance between spine body parts to report a huddle detection
             - huddle_speed (int): maximum speed to report a huddle detection
 
     Returns:
         - tag_df (pandas.DataFrame): table with traits as columns and frames as rows. Each
         value is a boolean indicating trait detection at a given time"""
 
-    hparams = get_hparameters(hparams)
+    params = get_hparameters(params)
     animal_ids = coordinates._animal_ids
     undercond = "_" if len(animal_ids) > 1 else ""
 
@@ -488,23 +501,23 @@ def rule_based_tagging(
 
     def onebyone_contact(bparts: List):
         """Returns a smooth boolean array with 1to1 contacts between two mice"""
-        nonlocal coords, animal_ids, hparams, arena_abs, arena
+        nonlocal coords, animal_ids, params, arena_abs, arena
 
         return deepof.utils.smooth_boolean_array(
             close_single_contact(
                 coords,
                 animal_ids[0] + bparts[0],
                 animal_ids[1] + bparts[-1],
-                hparams["close_contact_tol"],
+                params["close_contact_tol"],
                 arena_abs,
-                arena[2],
+                arena[1][1],
             )
         )
 
     def twobytwo_contact(rev):
         """Returns a smooth boolean array with side by side contacts between two mice"""
 
-        nonlocal coords, animal_ids, hparams, arena_abs, arena
+        nonlocal coords, animal_ids, params, arena_abs, arena
         return deepof.utils.smooth_boolean_array(
             close_double_contact(
                 coords,
@@ -512,13 +525,31 @@ def rule_based_tagging(
                 animal_ids[0] + "_Tail_base",
                 animal_ids[1] + "_Nose",
                 animal_ids[1] + "_Tail_base",
-                hparams["side_contact_tol"],
+                params["side_contact_tol"],
                 rev=rev,
                 arena_abs=arena_abs,
-                arena_rel=arena[2],
+                arena_rel=arena[1][1],
             )
         )
 
+    def overall_speed(speeds, _id, undercond):
+        bparts = [
+            "Center",
+            "Spine_1",
+            "Spine_2",
+            "Nose",
+            "Left_ear",
+            "Right_ear",
+            "Left_fhip",
+            "Right_fhip",
+            "Left_bhip",
+            "Right_bhip",
+            "Tail_base",
+        ]
+        array = speeds[[_id + undercond + bpart for bpart in bparts]]
+        avg_speed = np.nanmedian(array[1:], axis=1)
+        return np.insert(avg_speed, 0, np.nan, axis=0)
+
     if len(animal_ids) == 2:
         # Define behaviours that can be computed on the fly from the distance matrix
         tag_dict["nose2nose"] = onebyone_contact(bparts=["_Nose"])
@@ -541,23 +572,28 @@ def rule_based_tagging(
                     coords,
                     follower=_id,
                     followed=[i for i in animal_ids if i != _id][0],
-                    frames=hparams["follow_frames"],
-                    tol=hparams["follow_tol"],
+                    frames=params["follow_frames"],
+                    tol=params["follow_tol"],
                 )
             )
 
     for _id in animal_ids:
         tag_dict[_id + undercond + "climbing"] = deepof.utils.smooth_boolean_array(
-            climb_wall(arena_type, arena, coords, w / 100, _id + undercond + "Nose")
+            climb_wall(
+                arena_type,
+                arena,
+                coords,
+                params["climb_tol"],
+                _id + undercond + "Nose",
+            )
         )
-        tag_dict[_id + undercond + "speed"] = speeds[_id + undercond + "Center"]
+        tag_dict[_id + undercond + "speed"] = overall_speed(speeds, _id, undercond)
         tag_dict[_id + undercond + "huddle"] = deepof.utils.smooth_boolean_array(
             huddle(
                 coords,
                 speeds,
-                hparams["huddle_forward"],
-                hparams["huddle_spine"],
-                hparams["huddle_speed"],
+                params["huddle_forward"],
+                params["huddle_speed"],
                 animal_id=_id,
             )
         )
@@ -578,11 +614,15 @@ def tag_rulebased_frames(
     dims,
     undercond,
     hparams,
+    arena,
+    debug,
+    coords,
 ):
     """Helper function for rule_based_video. Annotates a given frame with on-screen information
     about the recognised patterns"""
 
     w, h = dims
+    arena, h, w = arena
 
     def write_on_frame(text, pos, col=(255, 255, 255)):
         """Partial closure over cv2.putText to avoid code repetition"""
@@ -600,9 +640,9 @@ def tag_rulebased_frames(
         if cond is None:
             cond = frame_speeds[animal_ids[0]] > frame_speeds[animal_ids[1]]
         if cond:
-            return 150, 150, 255
-        else:
             return 150, 255, 150
+        else:
+            return 150, 150, 255
 
     zipped_pos = list(
         zip(
@@ -614,7 +654,25 @@ def tag_rulebased_frames(
 
     if len(animal_ids) > 1:
 
-        cv2.line(frame, (100, 100), (110, 100), (255, 0, 0))
+        if debug:
+            # Print arena for debugging
+            cv2.ellipse(frame, arena[0], arena[1], arena[2], 0, 360, (0, 255, 0), 3)
+            # Print body parts for debuging
+            for bpart in coords.columns.levels[0]:
+                if not np.isnan(coords[bpart]["x"][fnum]):
+                    cv2.circle(
+                        frame,
+                        (int(coords[bpart]["x"][fnum]), int(coords[bpart]["y"][fnum])),
+                        radius=3,
+                        color=(
+                            (255, 0, 0)
+                            if bpart.startswith(animal_ids[0])
+                            else (0, 0, 255)
+                        ),
+                        thickness=-1,
+                    )
+            # Print frame number
+            write_on_frame("Frame " + str(fnum), corners["downleft"])
 
         if tag_dict["nose2nose"][fnum] and not tag_dict["sidebyside"][fnum]:
             write_on_frame("Nose-Nose", conditional_pos())
@@ -687,7 +745,8 @@ def rule_based_video(
     frame_limit: int = np.inf,
     recog_limit: int = 1,
     path: str = os.path.join("."),
-    hparams: dict = {},
+    params: dict = {},
+    debug: bool = False,
 ) -> True:
     """Renders a version of the input video with all rule-based taggings in place.
 
@@ -695,12 +754,14 @@ def rule_based_video(
         - tracks (list): list containing experiment IDs as strings
         - videos (list): list of videos to load, in the same order as tracks
         - coordinates (deepof.preprocessing.coordinates): coordinates object containing the project information
+        - debug (bool): if True, several debugging attributes (such as used body parts and arena) are plotted in
+        the output video
         - vid_index (int): index in videos of the experiment to annotate
         - fps (float): frames per second of the analysed video. Same as input by default
         - path (str): directory in which the experimental data is stored
         - frame_limit (float): limit the number of frames to output. Generates all annotated frames by default
         - recog_limit (int): number of frames to use for arena recognition (1 by default)
-        - hparams (dict): dictionary to overwrite the default values of the hyperparameters of the functions
+        - params (dict): dictionary to overwrite the default values of the hyperparameters of the functions
         that the rule-based pose estimation utilizes. Values can be:
             - speed_pause (int): size of the rolling window to use when computing speeds
             - close_contact_tol (int): maximum distance between single bodyparts that can be used to report the trait
@@ -709,7 +770,6 @@ def rule_based_video(
             - follow_tol (int): maximum distance between follower and followed's path during the last follow_frames,
             in order to report a detection
             - huddle_forward (int): maximum distance between ears and forward limbs to report a huddle detection
-            - huddle_spine (int): maximum average distance between spine body parts to report a huddle detection
             - huddle_speed (int): maximum speed to report a huddle detection
 
     Returns:
@@ -718,7 +778,7 @@ def rule_based_video(
     """
 
     # DATA OBTENTION AND PREPARATION
-    hparams = get_hparameters(hparams)
+    params = get_hparameters(params)
     animal_ids = coordinates._animal_ids
     undercond = "_" if len(animal_ids) > 1 else ""
 
@@ -755,12 +815,12 @@ def rule_based_video(
         try:
             if (
                 list(frame_speeds.values())[0] == -np.inf
-                or fnum % hparams["speed_pause"] == 0
+                or fnum % params["speed_pause"] == 0
             ):
                 for _id in animal_ids:
                     frame_speeds[_id] = tag_dict[_id + undercond + "speed"][fnum]
         except AttributeError:
-            if frame_speeds == -np.inf or fnum % hparams["speed_pause"] == 0:
+            if frame_speeds == -np.inf or fnum % params["speed_pause"] == 0:
                 frame_speeds = tag_dict["speed"][fnum]
 
         # Display all annotations in the output video
@@ -774,7 +834,10 @@ def rule_based_video(
             fnum,
             (w, h),
             undercond,
-            hparams,
+            params,
+            (arena, h, w),
+            debug,
+            coordinates.get_coords(center=False)[vid_name],
         )
 
         if writer is None:
@@ -784,7 +847,7 @@ def rule_based_video(
             writer.open(
                 vid_name + "_tagged.avi",
                 cv2.VideoWriter_fourcc(*"MJPG"),
-                hparams["fps"],
+                params["fps"],
                 (frame.shape[1], frame.shape[0]),
                 True,
             )
@@ -799,5 +862,4 @@ def rule_based_video(
 
 
 # TODO:
-#    - Relativise default contact parameters
-#    (right now they only work if 'arena_dims' is correctly set in the project object
+#    - Is border sniffing anything you might consider interesting?
