@@ -219,7 +219,7 @@ arena_dims = args.arena_dims
 batch_size = args.batch_size
 hypertun_trials = args.hpt_trials
 encoding_size = args.encoding_size
-exclude_bodyparts = tuple(args.exclude_bodyparts.split(","))
+exclude_bodyparts = args.exclude_bodyparts.split(",")
 gaussian_filter = args.gaussian_filter
 hparams = args.hyperparameters
 input_type = args.input_type
