@@ -19,6 +19,8 @@ import regex as re
 from copy import deepcopy
 from itertools import combinations, product
 from joblib import Parallel, delayed
+
+# from skimage.transform import hough_ellipse
 from sklearn import mixture
 from tqdm import tqdm
 from typing import Tuple, Any, List, Union, NewType
@@ -255,14 +257,14 @@ def align_trajectories(data: np.array, mode: str = "all") -> np.array:
     """Returns a numpy.array with the positions rotated in a way that the center (0 vector)
     and the body part in the first column of data are aligned with the y axis.
 
-        Parameters:
-            - data (3D numpy.array): array containing positions of body parts over time, where
-            shape is N (sliding window instances) * m (sliding window size) * l (features)
-            - mode (string): specifies if *all* instances of each sliding window get
-            aligned, or only the *center*
+    Parameters:
+        - data (3D numpy.array): array containing positions of body parts over time, where
+        shape is N (sliding window instances) * m (sliding window size) * l (features)
+        - mode (string): specifies if *all* instances of each sliding window get
+        aligned, or only the *center*
 
-        Returns:
-            - aligned_trajs (2D np.array): aligned positions over time"""
+    Returns:
+        - aligned_trajs (2D np.array): aligned positions over time"""
 
     angles = np.zeros(data.shape[0])
     data = deepcopy(data)
@@ -345,6 +347,154 @@ def smooth_mult_trajectory(series: np.array, alpha: float = 0.99) -> np.array:
     return smoothed_series
 
 
+def moving_average(time_series: pd.Series, N: int = 5):
+    """Fast implementation of a moving average function
+
+    Parameters:
+        - time_series (pd.Series): univariate time series to take the moving average of
+        - N (int): size of the convolution window used to compute the moving average
+
+    Returns:
+        -  moving_avg (pd.Series): univariate moving average over time_series"""
+
+    moving_avg = np.convolve(time_series, np.ones(N) / N, mode="same")
+
+    return moving_avg
+
+
+def mask_outliers(
+    time_series: pd.DataFrame,
+    likelihood: pd.DataFrame,
+    likelihood_tolerance: float,
+    lag: int,
+    n_std: int,
+    mode: str,
+):
+    """Returns a mask over the bivariate trajectory of a body part, identifying as True all detected outliers
+
+    Parameters:
+        - time_series (pd.DataFrame): bivariate time series representing the x, y positions of a single body part
+        - likelihood (pd.DataFrame): dataframe with likelihood data per body part as extracted from deeplabcut
+        - likelihood_tolerance (float): minimum tolerated likelihood, below which an outlier is called
+        - lag (int): size of the convolution window used to compute the moving average
+        - n_std (int): number of standard deviations over the moving average to be considered an outlier
+        - mode (str): if "and" (default) both x and y have to be marked in order to call an outlier.
+        If "or", one is enough
+
+    Returns
+        - mask (pd.DataFrame): bivariate mask over time_series. True indicates an outlier"""
+
+    moving_avg_x = moving_average(time_series["x"], lag)
+    moving_avg_y = moving_average(time_series["y"], lag)
+
+    residuals_x = time_series["x"] - moving_avg_x
+    residuals_y = time_series["y"] - moving_avg_y
+
+    outlier_mask_x = np.abs(residuals_x) > n_std * np.std(residuals_x)
+    outlier_mask_y = np.abs(residuals_y) > n_std * np.std(residuals_y)
+    outlier_mask_l = likelihood < likelihood_tolerance
+    mask = None
+
+    if mode == "and":
+        mask = (outlier_mask_x & outlier_mask_y) | outlier_mask_l
+    elif mode == "or":
+        mask = (outlier_mask_x | outlier_mask_y) | outlier_mask_l
+
+    return mask
+
+
+def full_outlier_mask(
+    experiment: pd.DataFrame,
+    likelihood: pd.DataFrame,
+    likelihood_tolerance: float,
+    exclude: str,
+    lag: int,
+    n_std: int,
+    mode: str,
+):
+    """Iterates over all body parts of experiment, and outputs a dataframe where all x, y positions are
+    replaced by a boolean mask, where True indicates an outlier
+
+    Parameters:
+        - experiment (pd.DataFrame): dataframe with time series representing the x, y positions of a every body part
+        - likelihood (pd.DataFrame): dataframe with likelihood data per body part as extracted from deeplabcut
+        - likelihood_tolerance (float): minimum tolerated likelihood, below which an outlier is called
+        - exclude (str): body part to exclude from the analysis (to concatenate with bpart alignment)
+        - lag (int): size of the convolution window used to compute the moving average
+        - n_std (int): number of standard deviations over the moving average to be considered an outlier
+        - mode (str): if "and" (default) both x and y have to be marked in order to call an outlier.
+        If "or", one is enough
+
+    Returns:
+        full_mask (pd.DataFrame): mask over all body parts in experiment. True indicates an outlier
+
+    """
+
+    body_parts = experiment.columns.levels[0]
+    full_mask = experiment.copy()
+
+    if exclude:
+        full_mask.drop(exclude, axis=1, inplace=True)
+
+    for bpart in body_parts:
+        if bpart != exclude:
+            mask = mask_outliers(
+                experiment[bpart],
+                likelihood[bpart],
+                likelihood_tolerance,
+                lag,
+                n_std,
+                mode,
+            )
+
+            full_mask.loc[:, (bpart, "x")] = mask
+            full_mask.loc[:, (bpart, "y")] = mask
+            continue
+
+    return full_mask
+
+
+def interpolate_outliers(
+    experiment: pd.DataFrame,
+    likelihood: pd.DataFrame,
+    likelihood_tolerance: float,
+    exclude: str = "",
+    lag: int = 5,
+    n_std: int = 3,
+    mode: str = "or",
+    limit: int = 15,
+):
+    """Marks all outliers in experiment and replaces them using a univariate linear interpolation approach.
+    Note that this approach only works for equally spaced data (constant camera acquisition rates).
+
+    Parameters:
+        - experiment (pd.DataFrame): dataframe with time series representing the x, y positions of a every body part
+        - likelihood (pd.DataFrame): dataframe with likelihood data per body part as extracted from deeplabcut
+        - likelihood_tolerance (float): minimum tolerated likelihood, below which an outlier is called
+        - exclude (str): body part to exclude from the analysis (to concatenate with bpart alignment)
+        - lag (int): size of the convolution window used to compute the moving average
+        - n_std (int): number of standard deviations over the moving average to be considered an outlier
+        - mode (str): if "and" both x and y have to be marked in order to call an outlier.
+        If "or" (default), one is enough
+        - limit (int): maximum of consecutive outliers to interpolate. Defaults to 15
+
+    Returns:
+        interpolated_exp (pd.DataFrame): Interpolated version of experiment
+    """
+
+    interpolated_exp = experiment.copy()
+    mask = full_outlier_mask(
+        experiment, likelihood, likelihood_tolerance, exclude, lag, n_std, mode
+    )
+
+    interpolated_exp[mask] = np.nan
+    interpolated_exp.interpolate(
+        method="linear", limit=limit, limit_direction="both", inplace=True
+    )
+
+    return interpolated_exp
+
+
 def recognize_arena(
     videos: list,
     vid_index: int,
@@ -383,7 +533,8 @@ def recognize_arena(
         if arena_type == "circular":
 
             # Detect arena and extract positions
-            arena = circular_arena_recognition(frame)[0]
+            arena = circular_arena_recognition(frame)
+
             if h is None and w is None:
                 w, h = frame.shape[0], frame.shape[1]
 
@@ -405,37 +556,36 @@ def circular_arena_recognition(frame: np.array) -> np.array:
         - circles (np.array): 3-element-array containing x,y positions of the center
         of the arena, and a third value indicating the radius"""
 
-    # Convert image to greyscale, threshold it, blur it and detect the biggest best fitting circle
-    # using the Hough algorithm
+    # Convert image to greyscale, threshold it, blur it and apply open-close operations
     gray_image = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
-    ret, thresh = cv2.threshold(gray_image, 50, 255, 0)
+    ret, thresh = cv2.threshold(gray_image, 80, 255, 0)
     frame = cv2.medianBlur(thresh, 9)
-    circle = cv2.HoughCircles(
-        frame,
-        cv2.HOUGH_GRADIENT,
-        1,
-        300,
-        param1=50,
-        param2=10,
-        minRadius=0,
-        maxRadius=0,
+    frame = cv2.morphologyEx(
+        frame, cv2.MORPH_CLOSE, cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (80, 80))
+    )
+    frame = cv2.morphologyEx(
+        frame, cv2.MORPH_OPEN, cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (80, 80))
     )
 
-    circles = []
+    # Find contours in the processed image
+    cnts, _ = cv2.findContours(frame, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
 
-    if circle is not None:
-        circle = np.uint16(np.around(circle[0]))
-        circles.append(circle)
+    # Detect the main ellipse containing the arena
+    ellipse_params = cv2.fitEllipse(cnts[0])
+    center_coordinates = tuple([int(i) for i in ellipse_params[0]])
+    axes_length = tuple([int(i) // 2 for i in ellipse_params[1]])
+    angle = ellipse_params[2]
 
-    return circles[0]
+    return center_coordinates, axes_length, angle
 
 
 def rolling_speed(
     dframe: pd.DatetimeIndex,
-    window: int = 5,
-    rounds: int = 10,
+    window: int = 3,
+    rounds: int = 3,
     deriv: int = 1,
     center: str = None,
+    shift: int = 2,
     typ: str = "coords",
 ) -> pd.DataFrame:
     """Returns the average speed over n frames in pixels per frame
@@ -469,12 +619,15 @@ def rolling_speed(
 
         features = 2 if der == 0 and typ == "coords" else 1
 
-        distances = np.concatenate(
-            [
-                np.array(dframe).reshape([-1, features], order="F"),
-                np.array(dframe.shift()).reshape([-1, features], order="F"),
-            ],
-            axis=1,
+        distances = (
+            np.concatenate(
+                [
+                    np.array(dframe).reshape([-1, features], order="C"),
+                    np.array(dframe.shift(shift)).reshape([-1, features], order="C"),
+                ],
+                axis=1,
+            )
+            / shift
         )
 
         distances = np.array(compute_dist(distances))
@@ -483,12 +636,10 @@ def rolling_speed(
                 original_shape[0],
                 (original_shape[1] // 2 if typ == "coords" else original_shape[1]),
             ],
-            order="F",
+            order="C",
         )
         distances = pd.DataFrame(distances, index=dframe.index)
         speeds = np.round(distances.rolling(window).mean(), rounds)
-        speeds[np.isnan(speeds)] = 0.0
-
         dframe = speeds
 
     speeds.columns = body_parts
